library ieee;
  use ieee.std_logic_1164.all;
  use ieee.numeric_std.all;

package clocking_lib is
  
  constant MAX_FREQUENCY: integer := 25000000;
  
  function GetPeriod(frequency: real) return real;
  function GetFrequency(period: real) return real;
  
  component bus_synchronizer is
    generic(
      DATAWIDTH     : integer := 1;
      AUTO_REQUEST  : boolean := true
    );
    port(
      -- TX signals @ tx_clk
      tx_clk        : in  std_logic;
      tx_rst        : in  std_logic;
      tx_data       : in  std_logic_vector(DATAWIDTH - 1 downto 0);
      tx_req        : in  std_logic; -- tx-request, single cycle
      tx_finished   : out std_logic; -- sync finished, single cycle
      
      -- RX signals @ rx_clk
      rx_clk        : in std_logic;
      rx_rst        : in std_logic;
      rx_data       : out std_logic_vector(DATAWIDTH - 1 downto 0);
      rx_done       : out std_logic  -- new data received, single cycle
    );
  end component;
  
  component reset_synchronizer is 
    port (
      clk       : in std_logic;
      reset_in  : in std_logic;
      reset_out : out std_logic
    );
  end component;
  
  component synchronizer is
    generic (
      DATAWIDTH     : integer := 1
    );
    port (
      clk           : in std_logic;
      cc_data_in    : in std_logic_vector(DATAWIDTH - 1 downto 0);
      cc_data_out   : out std_logic_vector(DATAWIDTH - 1 downto 0)     
    );
  end component;
  
end package clocking_lib;

package body clocking_lib is
  
  function GetPeriod(frequency: real) return real is
  begin
    if frequency > 0.0 then
      return 1.0 / frequency;
    end if;
    return 0.0;
  end function;
  
  function GetFrequency(period: real) return real is
  begin
    if period > 0.0 then
      return 1.0 / period;
    end if;
    return 0.0;
  end function;
  
end package body;

---------------------------------------------------------------------------
-- BUS CLOCK CROSSER                                                     --
---------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity bus_synchronizer is
  generic(
    DATAWIDTH     : integer := 1;
    AUTO_REQUEST  : boolean := true
  );
  port(
    -- TX signals @ tx_clk
    tx_clk        : in  std_logic;
    tx_rst        : in  std_logic;
    tx_data       : in  std_logic_vector(DATAWIDTH - 1 downto 0);
    tx_req        : in  std_logic; -- tx-request, single cycle
    tx_finished   : out std_logic; -- sync finished, single cycle
       
    -- RX signals @ rx_clk
    rx_clk        : in std_logic;
    rx_rst        : in std_logic;
    rx_data       : out std_logic_vector(DATAWIDTH - 1 downto 0);
    rx_done       : out std_logic  -- new data received, single cycle
  );
end bus_synchronizer;
   
architecture rtl of bus_synchronizer is
    
  signal tx_reg      : std_logic_vector(DATAWIDTH - 1 downto 0);
  signal req_sync    : std_logic_vector(2 downto 0);
  signal ack_sync    : std_logic_vector(2 downto 0);
  signal request     : std_logic;
  signal acknowledge : std_logic;
  signal tx_req_dmy  : std_logic;
    
  attribute ALTERA_INTERNAL_OPTION : string;
  attribute ALTERA_INTERNAL_OPTION OF RTL : ARCHITECTURE IS "-name SDC_STATEMENT ""set_false_path -from *acknowledge* -to *ack_sync* "";" &
                                                            "-name SDC_STATEMENT ""set_false_path -from *request*     -to *req_sync* "";" &
                                                            "-name SDC_STATEMENT ""set_false_path -from *tx_reg*      -to *rx_data* """;
    
begin
   
  tx_req_dmy <= tx_req when AUTO_REQUEST = false else not ack_sync(1) and not request;
  tx_finished <= not ack_sync(2) and ack_sync(1) when AUTO_REQUEST = false else '0';
  process (tx_clk, tx_rst) is
  begin
    if tx_rst = '1' then
      request  <= '0';
      tx_reg   <= (others => '0');
      ack_sync <= (others => '0');
    elsif rising_edge(tx_clk) then
      -- process the acknowledge signal
      ack_sync <= ack_sync(1 downto 0) & acknowledge;
          
      -- latch the data
      -- generate the request signal
      if tx_req_dmy = '1' and request = '0' and ack_sync(1) = '0' then
        tx_reg <= tx_data;
        request <= '1';
      end if;
         
      if ack_sync(1) = '1' then
        request <= '0';
      end if;
    end if;
  end process;
      
  rx_done <= not req_sync(2) and req_sync(1) when AUTO_REQUEST = false else '0';
  process (rx_clk, rx_rst) is
  begin
    if rx_rst = '1' then
      acknowledge <= '0';
      req_sync    <= (others => '0');
      rx_data     <= (others => '0');
    elsif rising_edge(rx_clk) then
      -- process the request signal
      req_sync <= req_sync(1 downto 0) & request;
          
      -- latch the data
      -- generate the acknowledge signal
      if req_sync(1) = '1' then
        rx_data <= tx_reg;
        acknowledge <= '1';
      else
        acknowledge <= '0';
      end if;
    end if;
  end process;
   
end rtl;
  
  
---------------------------------------------------------------------------
--  CLOCK CROSSER                                                        --
---------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity synchronizer is
  generic(
    DATAWIDTH     : integer := 1
  );
  port(
    clk           : in std_logic;
    cc_data_in    : in std_logic_vector(DATAWIDTH - 1 downto 0);
    cc_data_out   : out std_logic_vector(DATAWIDTH - 1 downto 0)     
  );
end synchronizer;
   
architecture rtl of synchronizer is
   
  attribute ALTERA_INTERNAL_OPTION : string;     
  attribute ALTERA_INTERNAL_OPTION OF RTL : ARCHITECTURE IS "-name SDC_STATEMENT ""set_false_path -from * -to *sync_data1_s* """;
     
  signal sync_data1 : std_logic_vector(DATAWIDTH - 1 downto 0);
  signal sync_data2 : std_logic_vector(DATAWIDTH - 1 downto 0);
  signal data       : std_logic_vector(DATAWIDTH - 1 downto 0);
   
begin
     
  process(clk)
  begin
    if rising_edge(clk) then
      sync_data1   <= cc_data_in;
      sync_data2   <= sync_data1;
      data         <= sync_data2;
      cc_data_out  <= sync_data1;
    end if;
  end process;
   
end rtl;

---------------------------------------------------------------------------
-- RESET CLOCK CROSSER                                                   --
---------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
  
entity reset_synchronizer is 
  port (
    -- inputs:
    clk       : IN STD_LOGIC;
    reset_in  : IN STD_LOGIC;
    -- outputs:
    reset_out : OUT STD_LOGIC
  );
end entity reset_synchronizer;


architecture rtl of reset_synchronizer is
  signal sync_data :  STD_LOGIC;
  signal sync_reset :  std_logic;

  attribute ALTERA_INTERNAL_OPTION : string;     
  attribute ALTERA_INTERNAL_OPTION OF RTL : ARCHITECTURE IS "-name SDC_STATEMENT ""set_false_path -from * -to *sync_reset* ""; " & 
                                                            "-name SDC_STATEMENT ""set_false_path -from * -to *sync_data* """;

begin

  process (clk, reset_in)
  begin
    if reset_in = '1' then
      sync_data <= '0';
    elsif rising_edge(clk) then
      sync_data <= '1';
    end if;
  end process;
                                         
  process (clk, reset_in)
  begin
    if reset_in = '1' then
      sync_reset <= '0';
    elsif rising_edge(clk) then
      sync_reset <= sync_data;
    end if;
  end process;
    
  reset_out <= not sync_reset;
end rtl;
