library IEEE;
  use IEEE.std_logic_1164.all;
  use IEEE.numeric_std.all;
  use IEEE.std_logic_unsigned.all;
  
library spi_lib;
  use spi_lib.spi_lib.all;  
  
library clocking_lib;
  use clocking_lib.clocking_lib.all;
  
entity spi is
  generic (
    SPI_OWN_ADDRESS_G : std_logic_vector(2 downto 0) := "110"
  );
  port (
    -- Interface to/from main CPU
    clk_i               : in  std_logic;
    rst_i               : in  std_logic;
    spi_wr_data_i       : in  spi_data_t;
    spi_wr_done_o       : out std_logic;
    spi_rd_data_o       : out spi_data_t;
    -- Physical SPI interface
    spi_clk_i           : in  std_logic;
    spi_cs_i            : in  std_logic_vector(2 downto 0);
    spi_miso_o          : out std_logic;
    spi_mosi_i          : in  std_logic
  );
end entity;

architecture rtl of spi is

  -- SPI READ MODULE
  signal spi_shifter    : std_logic_vector(SPI_DATA_WIDTH_C - 1 downto 0);
  signal spi_data_cnt   : unsigned(7 downto 0);
  signal spi_cs_s       : std_logic_vector(2 downto 0);
  signal spi_clk_s      : std_logic;
  signal spi_mosi_s     : std_logic;
  signal spi_clk_ss     : std_logic;
  signal spi_clk_pos    : std_logic;
  signal spi_clk_neg    : std_logic;

  -- SPI WRITE MODULE
  signal spi_wr_shifter : std_logic_vector(SPI_DATA_WIDTH_C - 1 downto 0);
  signal spi_wr_valid   : std_logic;
  signal spi_miso_s     : std_logic;
  
begin

  -- we sample the spi-clk signal to shift in the spi data in the right clock domain from the beginning...
  -- NOTICE: the spi_clk signal is delayed by two clk_i cycles! The other spi signals are, however, NOT
  -- delayed. This might make problems! We assume that the oversampling clock is fast enough.
  -- We simply sample all the data signal as well!
  spi_clk_sync_inst: synchronizer
  port map (clk_i, cc_data_in(0) => spi_clk_i, cc_data_out(0) => spi_clk_s);

  -- see comment above
  spi_data_sync_inst: synchronizer
  port map (clk_i, cc_data_in(0) => spi_mosi_i, cc_data_out(0) => spi_mosi_s);
  
  spi_data_out_inst: synchronizer
  port map (clk_i, cc_data_in(0) => spi_wr_shifter(0), cc_data_out(0) => spi_miso_s);

  -- we simply use a two-stage synchronizer for the 3-bit CS signal. Usually, this is a bad idea (because
  -- the resulting 3 bits might not be correlated to each other any more (image one bit being a little early,
  -- the other being a little late).
  -- however, when we actually need the spi_cs_s signal, an entire SPI-frame has been transmitted with the
  -- CS signal being stable long enough. We can be sure that it has been synchronized correctly by then.
  spi_cs_sync_inst: synchronizer
  generic map (3)
  port map (clk_i, spi_cs_i, spi_cs_s);

  spi_clk_pos <= '1' when spi_clk_ss /= spi_clk_s and spi_clk_s = '1' else '0';
  spi_clk_neg <= '1' when spi_clk_ss /= spi_clk_s and spi_clk_s = '0' else '0';
  spi_miso_o <= spi_miso_s when spi_cs_s = SPI_OWN_ADDRESS_G else 'Z';
  
  spi_input_p: process (clk_i, rst_i)
  begin
    if rst_i = '1' then
      spi_shifter  <= (others => '0');
      spi_data_cnt <= (others => '0');
      spi_clk_ss   <= '0';
      spi_rd_data_o.valid <= '0';
      spi_rd_data_o.data <= (others => '0');
    elsif rising_edge(clk_i) then
      spi_rd_data_o.valid <= '0';
      spi_clk_ss <= spi_clk_s;
      -- on rising spi_clk_i edge...
      if spi_clk_neg = '1' then
        -- we assume LSB is sent first
        spi_shifter <= spi_mosi_s & spi_shifter(SPI_DATA_WIDTH_C - 2 downto 0);
        if spi_data_cnt /= SPI_DATA_WIDTH_C - 1 then
          spi_data_cnt <= spi_data_cnt + 1;
        else
          spi_data_cnt <= (others => '0');
          if spi_cs_s = SPI_OWN_ADDRESS_G then
            spi_rd_data_o.data <= spi_shifter;
            spi_rd_data_o.valid <= '1';
          end if;
        end if;
      end if;
    end if;
  end process;
  
  spi_output_p: process (clk_i, rst_i) is
  begin
    if rst_i = '1' then
      spi_wr_shifter <= (others => '0');
      spi_wr_done_o  <= '0';
      spi_wr_valid   <= '0';
    elsif rising_edge(clk_i) then
      spi_wr_done_o <= '0';
      -- assign the new word which is to be transmitted (only when no current transmission is scheduled => check spi_wr_valid)
      if spi_wr_data_i.valid = '1' and spi_wr_valid = '0' then
        spi_wr_valid <= '1';
        spi_wr_shifter <= spi_wr_data_i.data;
      end if;
      
      if spi_clk_pos = '1' then
        spi_wr_shifter <= "0" & spi_wr_shifter(SPI_DATA_WIDTH_C - 1 downto 1);
        if spi_data_cnt = SPI_DATA_WIDTH_C - 1 and spi_cs_s = SPI_OWN_ADDRESS_G then
          spi_wr_done_o <= '1';
          spi_wr_valid <= '0';
        end if;
      end if;
    end if;
  end process;
  
end architecture;
