library IEEE;
  use IEEE.std_logic_1164.all;
  use IEEE.numeric_std.all;
  use IEEE.std_logic_unsigned.all;

library clocking_lib;
  use clocking_lib.clocking_lib.all;

library spi_lib;
  use spi_lib.spi_lib.all;

entity top is
  port (
    dbg: out std_logic := '0'
  );
end entity;

architecture rtl of top is

  component spi is
    generic (
      SPI_OWN_ADDRESS_G : std_logic_vector(2 downto 0) := "110"
    );
    port (
      -- Interface to/from main CPU
      clk_i               : in  std_logic;
      rst_i               : in  std_logic;
      spi_wr_data_i       : in  spi_data_t;
      spi_wr_done_o       : out std_logic;
      spi_rd_data_o       : out spi_data_t;
      -- Physical SPI interface
      spi_clk_i           : in  std_logic;
      spi_cs_i            : in  std_logic_vector(2 downto 0);
      spi_miso_o          : out std_logic;
      spi_mosi_i          : in  std_logic
    );
  end component;

  type state_t is (IDLE, SEND, DATA);
  
  signal clk, rst                 : std_logic;
  
  signal reset_s                  : std_logic;
  signal spi_miso, spi_mosi       : std_logic;
  signal spi_clk, spi_wr_done     : std_logic := '0';
  signal spi_cs                   : std_logic_vector(2 downto 0);
  signal spi_rd_data, spi_wr_data : spi_data_t;

  signal state                    : state_t;
  signal clk_cnt                  : unsigned(31 downto 0) := (others => '0');
  signal idle_cnt                 : unsigned(7 downto 0);
  signal send_done                : std_logic := '0';
  signal send_data                : std_logic_vector(31 downto 0);
  signal spi_clk_cnt              : unsigned(4 downto 0);
  
begin

  dbg <= XOR_ALL(spi_rd_data.data) xor spi_rd_data.valid xor spi_miso;
  
  -- synchronize the asynchronous reset with our clock domain.
  reset_sync_inst: reset_synchronizer 
  port map (clk, rst, reset_s);

  spi_inst: spi
  generic map (
    SPI_OWN_ADDRESS_G => "101"
  )
  port map (
    clk_i         => clk,
    rst_i         => rst,
    spi_wr_data_i => spi_wr_data,
    spi_wr_done_o => spi_wr_done,
    spi_rd_data_o => spi_rd_data,
    spi_clk_i     => spi_clk,
    spi_cs_i      => spi_cs,
    spi_miso_o    => spi_miso,
    spi_mosi_i    => spi_mosi
  );
  
  -- RESET
  process is
  begin
    rst <= '1';
    wait for 500 ns;
    rst <= '0';
    wait;
  end process;
  
  -- fast CLK
  process is
  begin
    clk <= '1';
    wait for 5 ns;
    clk <= '0';
    wait for 5 ns;
    clk_cnt <= clk_cnt + 1;
  end process;
  
  -- slow SPI CLK
  process is
  begin
    wait until state = SEND;
    for i in 0 to 31 loop
      spi_clk <= '1';
      wait for 50 ns;
      spi_clk <= '0';
      wait for 50 ns;
    end loop;
    
    send_done <= '1';
    wait for 50 ns;
    send_done <= '0';
  end process;
  
  -- SPI send simulation
  process (rst, spi_clk) is
  begin
    if rst = '1' then
      spi_mosi <= '0';
      spi_clk_cnt <= (others => '0');
    elsif spi_clk'event and spi_clk = '1' then
      spi_clk_cnt <= spi_clk_cnt + 1;
      spi_mosi <= send_data(CONV_INTEGER(spi_clk_cnt));
    end if;
  end process;
  
  -- main state machine
  process (rst, clk) is
  begin
    if rst = '1' then
      state <= IDLE;
      spi_wr_data.data <= (others => '0');
      spi_wr_data.valid <= '0';
      idle_cnt <= (others => '0');
      spi_cs <= "010";
      send_data <= (others => '0');
    elsif rising_edge(clk) then
      spi_wr_data.valid <= '0';
      idle_cnt <= idle_cnt + 1;
      
      case state is
        when IDLE =>
          if idle_cnt(idle_cnt'left) = '1' then
            state <= SEND;
          end if;
          
        when SEND =>
          if send_done = '1' then
            state <= DATA;
          end if;
          
        when DATA =>
          send_data <= not std_logic_vector(clk_cnt);
          spi_wr_data.data <= std_logic_vector(clk_cnt);
          spi_wr_data.valid <= '1';
          idle_cnt <= (others => '0');
          spi_cs <= not spi_cs;
          state <= IDLE;
          
        when others =>
          null;
          
      end case;
    end if;
  end process;
  
end architecture;